"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
class UrlNode {
    constructor() {
        this.placeholder = true;
        this.children = new Map();
        this.slugName = null;
    }
    hasSlug() {
        return this.slugName != null;
    }
    insert(urlPath) {
        this._insert(urlPath.split('/').filter(Boolean));
    }
    smoosh() {
        return this._smoosh();
    }
    _smoosh(prefix = '/') {
        const childrenPaths = [...this.children.keys()].sort();
        if (this.hasSlug()) {
            childrenPaths.splice(childrenPaths.indexOf('[]'), 1);
        }
        const routes = childrenPaths
            .map(c => this.children.get(c)._smoosh(`${prefix}${c}/`))
            .reduce((prev, curr) => [...prev, ...curr], []);
        if (this.hasSlug()) {
            routes.push(...this.children.get('[]')._smoosh(`${prefix}[${this.slugName}]/`));
        }
        if (!this.placeholder) {
            routes.unshift(prefix === '/' ? '/' : prefix.slice(0, -1));
        }
        return routes;
    }
    _insert(urlPaths, slugNames = []) {
        if (urlPaths.length === 0) {
            this.placeholder = false;
            return;
        }
        // The next segment in the urlPaths list
        let nextSegment = urlPaths[0];
        // Check if the segment matches `[something]`
        if (nextSegment.startsWith('[') && nextSegment.endsWith(']')) {
            // Strip `[` and `]`, leaving only `something`
            const slugName = nextSegment.slice(1, -1);
            // If the specific segment already has a slug but the slug is not `something`
            // This prevents collisions like:
            // pages/[post]/index.js
            // pages/[id]/index.js
            // Because currently multiple dynamic params on the same segment level are not supported
            if (this.hasSlug() && slugName !== this.slugName) {
                // TODO: This error seems to be confusing for users, needs an err.sh link, the description can be based on above comment.
                throw new Error('You cannot use different slug names for the same dynamic path.');
            }
            if (slugNames.indexOf(slugName) !== -1) {
                throw new Error(`You cannot have the same slug name "${slugName}" repeat within a single dynamic path`);
            }
            slugNames.push(slugName);
            // slugName is kept as it can only be one particular slugName
            this.slugName = slugName;
            // nextSegment is overwritten to [] so that it can later be sorted specifically
            nextSegment = '[]';
        }
        // If this UrlNode doesn't have the nextSegment yet we create a new child UrlNode
        if (!this.children.has(nextSegment)) {
            this.children.set(nextSegment, new UrlNode());
        }
        this.children.get(nextSegment)._insert(urlPaths.slice(1), slugNames);
    }
}
function getSortedRoutes(normalizedPages) {
    // First the UrlNode is created, and every UrlNode can have only 1 dynamic segment
    // Eg you can't have pages/[post]/abc.js and pages/[hello]/something-else.js
    // Only 1 dynamic segment per nesting level
    // So in the case that is test/integration/dynamic-routing it'll be this:
    // pages/[post]/comments.js
    // pages/blog/[post]/comment/[id].js
    // Both are fine because `pages/[post]` and `pages/blog` are on the same level
    // So in this case `UrlNode` created here has `this.slugName === 'post'`
    // And since your PR passed through `slugName` as an array basically it'd including it in too many possibilities
    // Instead what has to be passed through is the upwards path's dynamic names
    const root = new UrlNode();
    // Here the `root` gets injected multiple paths, and insert will break them up into sublevels
    normalizedPages.forEach(pagePath => root.insert(pagePath));
    // Smoosh will then sort those sublevels up to the point where you get the correct route definition priority
    return root.smoosh();
}
exports.getSortedRoutes = getSortedRoutes;
